//WARNING comment syntax is important,  if unsure, check out http://www.doxygen.org/manual.html

//---------------------------------------------------------------------------
//! \file drivers.c
//!
//! \brief  Garage project. Device drivers
//!
//! Low-level IO drivers & hardware virtualisation for the garage door control.
//! This code can be used with any high-level strategy.
//!
//! \author David R Brooks
//!
//! \version 18/06/2007 DRB inital design and coding
//!
//! \attention           CPU: Atmel ATmega08
//! \note                Assumes 8MHz clock (internal)
//!
//---------------------------------------------------------------------------
// 2007.09.23 17:54  [YYYY.MM.DD] Last edited (CodeWright)

//-Included files -----------------------------------------------------------
#include "typedefn.h"
#include "interrupt.h"
#include "drivers.h"
#include "main.h"
#include "vt100.h"
#include "serial.h"

//-Globals ------------------------------------------------------------------

//-Locals -------------------------------------------------------------------

//-Local prototypes ---------------------------------------------------------
//! \cond SEE_PROTOTYPE

// [insert local protoypes]

//! \endcond

//-Local Functions ----------------------------------------------------------
// [insert local functions]

//-Global Functions ---------------------------------------------------------

//--StartHardware--------------------------------------------------------------
//! Hardware setups
//-----------------------------------------------------------------------------
//! Initialise all hardware resources, in the correct order
//! 
//-----------------------------------------------------------------------------
void StartHardware(void)
{
    IoPortInit();
    TimebaseInit(CFG_UPDATE_MS);
    CommsInit(CFG_COMMRATE_BD);
    AdcInit();                      // Must follow IoPortInit()
	
    wait(CFG_STARTDEL_US);          // Let the comms. start up
}

//--IoPortInit-----------------------------------------------------------------
//! Set up pin directions, etc.
//-----------------------------------------------------------------------------
//! Set up the default pin values & directions
//! D7 (learn) is open drain: bit is fixed = 0, control by switching direction
//! 
//! \note             Reset defines all port & DDR regs. as zero
//! 
//-----------------------------------------------------------------------------
void IoPortInit(void)
{
    PORTB = 0;
    DDRB  = INIT_DDR_B;
    PORTC = 0;
    DDRC  = INIT_DDR_C;
    PORTD = 0;
    DDRD  = INIT_DDR_D;
}

//--AdcInit--------------------------------------------------------------------
//! ADC initialisation (XTAL / 128)
//-----------------------------------------------------------------------------
//! Set up the ADC, waiting to start
//! 
//! \pre              Initialise IO ports first
//! \post             Write 1 to ADCSRA.ADSC to start conversion
//! 
//-----------------------------------------------------------------------------
void AdcInit(void)
{                       // Clock scaled by 128
    ADCSRA = BV(ADEN)  | BV(ADIE) | BV(ADPS2) | BV(ADPS1) | BV(ADPS0);
}

//--TimebaseInit---------------------------------------------------------------
//! Set up the main step timer (20mS)
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param period_u2  Update period (mS)
//!
//! \note             System clock = 8MHz, prescaler = 1024 giving a nominal 8kHz timebase
//! 
//-----------------------------------------------------------------------------
void TimebaseInit(uint2 period_u2)
{
    uint2 period_ms_u2 = period_u2 * CLKSCALE;  // Nominal clock is 8kHz

    TCCR1A = 0;                                 // Mode 4 (CTC, OCR1A), no output pins
    TCCR1B = BV(WGM12) | BV(CS12) | BV(CS10);   // Prescaler: CLK/1024
    OCR1A  = period_ms_u2;                      // Timer limit
    TIMSK |= BV(OCIE1A);                        // Interrupt on
}

//--IsRadio--------------------------------------------------------------------
//! Get state of current radio input
//-----------------------------------------------------------------------------
//! The radio input is determined by the config. reord
//! 
//! \return           State of selected line
//! 
//! \pre              Line selection comes from the Config. record
//! 
//-----------------------------------------------------------------------------
bool1 IsRadio(void)
{
    uint1 rx;

    switch(CFG_RXCHAN_E)
    {
    case CHAN_A:
        rx = RX_A;
        break; 
    case CHAN_B:
        rx = RX_B;
        break; 
    case CHAN_C:
        rx = RX_C;
        break; 
    case CHAN_D:
        rx = RX_D;
        break; 
    case CHAN_ALL:
        rx = RX_ALL;
        break; 
    default:
        return FALSE;
    }
    return (rx != 0);
}
