//WARNING comment syntax is important,  if unsure, check out http://www.doxygen.org/manual.html

//---------------------------------------------------------------------------
//! \file main.c
//!
//! \brief  Garage-door control: mainline code
//!
//! Main-line code: supervisory functions
//!     Version:    1.04         
//!     Processor:  Atmel ATMega8
//!     Compiler:   GCC          
//!
//! \author David R Brooks
//!
//! \version 14/07/2007 DRB inital design and coding
//!          25/08/2007     fixed bugs. Stable release
//!
//---------------------------------------------------------------------------
// 2007.10.06 14:38  [YYYY.MM.DD] Last edited (CodeWright)

// Note: The corrugation sensor is very unreliable: no development has been 
//       done on this function (stall detect).

//-Included files -----------------------------------------------------------
#include "main.h"
#include "vt100.h"
#include "serial.h"

//-Local Symbols ------------------------------------------------------------

//-Globals ------------------------------------------------------------------
// [insert global variables]

//-Locals -------------------------------------------------------------------

E_CSTATE state    = s_idle;         // Main state machine

// Scan order for ADC channels (must match E_INPUT)
const uint1 adc_chan_u1[NADCS] = {LIMTOP_ADC, LIMBTM_ADC};

bool1 timerflag = FALSE;	// Timer has run: do an update cycle
uint2 timeout   = 0;
uint2 count     = 0;

// Internal output-drive signals (so we can read them here)
bool1 i_down     = FALSE;
bool1 i_start    = FALSE;
bool1 i_power    = FALSE;
bool1 i_learn    = FALSE;

// Bitmaps of sensor inputs
uint1 InputsRaw_u1 		= 0;	// Raw switch inputs
uint1 InputsFiltered_u1 = 0;	// Filtered inputs
uint1 InputsEdge_u1		= 0;	// Edge detectors

uint2 RawAdcs[8] = {0};			// Raw ADCs for monitoring
uint2 LowestHigh, HighestLow;	// Limits for logging

//-Local prototypes ---------------------------------------------------------
//! \cond SEE_PROTOTYPE

static bool1 StopNow(void);
static void  ShowInputs(void);
static void  DriveLeds(void);

//! \endcond

//-Local Functions ----------------------------------------------------------

//-Global Functions ---------------------------------------------------------

//--SIGNAL(SIG_OUTPUT_COMPARE1A)-----------------------------------------------
//! Timer interrupt
//-----------------------------------------------------------------------------
//! Main timer interrupt. Set the hardware-watchdog bit, & trigger the ADC.
//! 
//-----------------------------------------------------------------------------
SIGNAL(SIG_OUTPUT_COMPARE1A)
{
    SETWATCHDOG(1);                     	// Hit the watchdog line
    SETBIT(InputsRaw_u1, S_ERROR_TOP, 0);	// No errors yet
    SETBIT(InputsRaw_u1, S_ERROR_BTM, 0);
    START_ADC(adc_chan_u1[0]);				// Kick off 1st ADC conversion cycle
}

//--SIGNAL---------------------------------------------------------------------
//! ADC operation-complete interrupt
//-----------------------------------------------------------------------------
//! The ADC is triggered by the main timer. This fn. converts 2 channels
//-----------------------------------------------------------------------------
SIGNAL(SIG_ADC)
{
    static uint1 chanptr_u1 = 0;        // Ptr. to following arrays
    uint2  adcb = ADC;                  // Grab the ADC input
    
    RawAdcs[chanptr_u1] = adcb;			// Value for monitoring
    
                                        // Check against limits
    if((adcb <= ADCRAW(CFG_ADCMAX_MV)) && (adcb >= ADCRAW(CFG_ADCHIGH_MV)))	   
    {
        SETBIT(InputsRaw_u1, chanptr_u1, ADCSENSE);		// Good, High value
        if(adcb < LowestHigh)
        {
        	LowestHigh = adcb;
        }
    }
    else
    if((adcb <= ADCRAW(CFG_ADCLOW_MV)) && (adcb >= ADCRAW(CFG_ADCMIN_MV)))
    {
        SETBIT(InputsRaw_u1, chanptr_u1, !ADCSENSE);	// Good, Low value
        if(adcb > HighestLow)
        {
        	HighestLow = adcb;
        }
    }
    else
    {
        SETBIT(InputsRaw_u1, chanptr_u1 + S_ERROR_TOP, 1);	// Faulty value
    	outs_P(PSTR("\nADC fault:"));
        outn(chanptr_u1, 2, 10);
        outn(adcb, 5, 10);
    }
    
    chanptr_u1++;

    if(chanptr_u1 < NADCS)              // More to go?
    {
        START_ADC(adc_chan_u1[chanptr_u1]); // Trigger next conversion
    }
    else
    {
        chanptr_u1 = 0;                 // Ready for next cycle
        timerflag  = TRUE;              // Run the main control logic
    }
}

//--ExternalInputs-------------------------------------------------------------
//! Debounce the switch inputs
//-----------------------------------------------------------------------------
//! When this is called, the raw values have been updated from the inputs
//-----------------------------------------------------------------------------
void ExternalInputs(void)
{
    static uint1 inputs_delayed_u1 = 0;		// Delayed values
    static uint1 inputs_filtdel_u1 = 0;		// Filtered, delayed
    
    uint1 x, y, z;
    
	SETBIT(InputsRaw_u1, S_RADIO,  IsRadio());  // Get the RX signals
	SETBIT(InputsRaw_u1, S_BUTTON, BUTTON);	    // Pushbutton
	SETBIT(InputsRaw_u1, S_INDEX,  INDEX);	    // Shaft sensor

    									// Combinatorial logic
    x = InputsRaw_u1 ^ inputs_delayed_u1;	// Noise detector
    y = InputsRaw_u1 & ~x;
    z = InputsFiltered_u1 & x;				// Mux: (raw == delay) ? raw : old
    									// Combinatorial outputs
    InputsFiltered_u1 = y | z;									// Filtered value
    InputsEdge_u1     = InputsFiltered_u1 ^ inputs_filtdel_u1;	// Any changes?
    									// Internal registers
    inputs_delayed_u1 = InputsRaw_u1;							// Delayed inputs
    inputs_filtdel_u1 = InputsFiltered_u1;
}

//--StopNow--------------------------------------------------------------------
//! Return currently-relevant limit switch (ie stop here)
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \return           TRUE if reached the limit
//! 
//-----------------------------------------------------------------------------
static bool1 StopNow(void)
{
    return((i_down) ? SENSOR(S_LIM_BTM) : SENSOR(S_LIM_TOP));
}

//--StateMachine---------------------------------------------------------------
//! The main state machine
//-----------------------------------------------------------------------------
//! The state machine handles all control functions, and interprets user
//! actions on the radio or pushbutton.
//! 
//-----------------------------------------------------------------------------
void StateMachine(void)
{
    E_CSTATE nextstate = state;
    PGM_P    reason_for_change = NULL;

    timeout += CFG_UPDATE_MS;            // Hard limit on operations
    count   += CFG_UPDATE_MS;            // Delay timer

	// Global safety-checks
    if(timeout  > CFG_MOTIONTIMEOUT_MS)         // Running too long
    {
    	reason_for_change = PSTR("Timed out");
        state = s_catch;
    }

    if(SENSOR(S_LIM_TOP) & SENSOR(S_LIM_BTM)) 	// Both limit switches active
    {
    	reason_for_change = PSTR("Both limit switches ON");
        state = s_catch;
    }

    if(SENSOR(S_ERROR_TOP) || SENSOR(S_ERROR_BTM))  // Switch wiring fault
    {
    	reason_for_change = PSTR("Switch wiring fault");
        state = s_catch;
    }

    switch(state)
    {
    // Wait here, until something happens
	case s_idle:
        timeout  = 0;
        OUT_POWER(FALSE);               // Be sure everything's off
        OUT_DOWN(FALSE); 
        OUT_START(FALSE);
        
        if(SENSOR(S_RADIO) || SENSOR(S_BUTTON)) // Start up...
        {
        	LowestHigh = 10000;			// Initialise the ADC-test extents
            HighestLow = 0;
            
        	// Choose direction to move
	        if(InputsRaw_u1 & BV(S_LIM_BTM)) 	// At bottom: must go up
	        {
	            i_down = FALSE;
	        }
	        else
	        if(InputsRaw_u1 & BV(S_LIM_TOP)) 	// At top: must go down
	        {
	            i_down = TRUE;
	        }
	        else                        		// Intermediate: reverse last motion
	        {
	            i_down = !i_down;
	        }

    		reason_for_change = (i_down) ? PSTR("Start down") : PSTR("Start up");
            nextstate = s_pick;					// To next state
        }
        break;

	// Starting a cycle: choose the direction, set the small relays
    // (start & direction). Wait for them to settle
    case s_pick:
        OUT_DOWN(i_down);           			// Set dirn. & start-mode
        OUT_START(TRUE);

        if(count >= CFG_PICKDELAY_MS)
        {
            OUT_POWER(TRUE);
    		reason_for_change = PSTR("Relays settled");
            nextstate = s_start;
        } 
        break;

	// Start the motor: let it pick up speed
    case s_start:
        if(StopNow())
        {
            OUT_POWER(FALSE);
    		reason_for_change = PSTR("Reached limit (start)");
            nextstate = s_end;
        }
        else
        if(count >= CFG_STARTDELAY_MS)
        {
            OUT_START(FALSE);
    		reason_for_change = PSTR("Up to speed");
            nextstate = s_index;
        } 
        break;
        
    // Wait for first index pulse (may be delayed, if the clutch
    // needs to engage first)
    case s_index:
        if(StopNow())				// Reached the end limit
        {
            OUT_POWER(FALSE);
    		reason_for_change = PSTR("Reached limit (clutch)");
            nextstate = s_end;
        }
    	else if(EDGE(S_INDEX))		// Index pulse seen (clucth is engaged)
        {
    		reason_for_change = PSTR("Index seen");
            nextstate = s_run;
        }
        else if(count >= CFG_CLUTCHTIMEOUT_MS)
        {
    		reason_for_change = PSTR("Clutch missed");
            nextstate = s_catch;
        }
    	break;

	// Live running: until the stop switch
    case s_run:
        if(StopNow())
        {
            OUT_POWER(FALSE);
    		reason_for_change = PSTR("Reached limit (run)");
            nextstate = s_end;
        }
        else if(EDGE(S_INDEX))
        {
        	count = 0;			// Reset the stall counter
        }
        else if(count >= CFG_CLUTCHTIMEOUT_MS)
        {
    		reason_for_change = PSTR("Drive stalled");
            nextstate = s_catch;
        }
    	break;

	// Reached the final limit switch: stop
    case s_end:                 // Done, or errors
        OUT_POWER(FALSE);
        if(!(SENSOR(S_RADIO) || SENSOR(S_BUTTON)))	// Wait until the controls are released
        {
			reason_for_change = PSTR("Finished");
            outs_P(PSTR("\nADC max low:"));
            outn(HighestLow, 5, 10);
            outs_P(PSTR(" min. high:"));
            outn(LowestHigh, 5, 10);
	        nextstate = s_idle;
    	}
    	break;
    
    // Illegal states: treat as fatal errors    
    default:
        OUT_POWER(FALSE);
		reason_for_change = PSTR("Illegal state");
        nextstate = s_catch;
    	break;
        
    // Various fatal errors    
    case s_catch:
        timeout = 0;
        OUT_POWER(FALSE);
        OUT_DOWN(FALSE); 
        OUT_START(FALSE);
        if(!(SENSOR(S_RADIO) || SENSOR(S_BUTTON)))
        {							// Wait if the signal is on
            nextstate = s_fault;
        }
        break;
        
    // Various fatal errors    
    case s_fault:
        timeout = 0;
        if(SENSOR(S_BUTTON))
        {							// User must signal (button) to exit
            nextstate = s_exit;
        }
        break;
        
    // Various fatal errors    
    case s_exit:
        timeout = 0;
        if(!(SENSOR(S_RADIO) || SENSOR(S_BUTTON)))
        {							// And that signal must clear
            nextstate = s_idle;
        }
        break;
    }

	// Actions when we change state
    if((state != nextstate)||((InputsEdge_u1 & ~BV(S_INDEX)) != 0))	// State/input change
    {
    	outc('\n');
        outn(timeout, 5, 10);
        outc(' ');
        ShowInputs();

    	if(reason_for_change)			// Explain why
        {
        	outc(' ');
            outs_P(reason_for_change);
        }

        state = nextstate;
        count = 0;
    }
}

static void ShowInputs(void)
{
	static char pinpname[][8] flash = SIGNAMES;
    int i;
    for(i=0; i<8; i++)
    {
    	uint1 mask_u1 = BV(i);
        if(InputsEdge_u1 & mask_u1)		// Signal has changed
        {
        	outs_P(pinpname[i]);
            outc(':');
            outs_P((InputsFiltered_u1 & mask_u1) ? PSTR("On") : PSTR("Off"));
            outc(' ');
        }
    }	
}

//--DriveLeds------------------------------------------------------------------
//! Drive LEDs from the current state
//-----------------------------------------------------------------------------
//! 
//-----------------------------------------------------------------------------
static void  DriveLeds(void)
{
	static uint2 ctr    = 0;
    static bool1 toggle = FALSE;	// Used for blink signals
    
    bool1 button = FALSE;			// Default LED states
    bool1 red    = FALSE;
    bool1 green  = FALSE;
    
    switch(state)
    {
	case s_end:			    // Stop moving
	case s_idle:			// Ready
    	ctr += CFG_UPDATE_MS;
        button = toggle;
        if(ctr >= ((toggle) ? CFG_LED_IDLE_ON_MS : CFG_LED_IDLE_OFF_MS))
        {
	        ctr = 0;
	        toggle = !toggle;
        }
    	break;

	case s_pick:			// Set the start/dirn. relays
	case s_start:			// Motor ON, start active
    	red = TRUE;
    	break;

	case s_index:			// Waiting for 1st Index
    	button = TRUE;
    	break;

	case s_run:				// Running...
    	button = TRUE;
        red    = TRUE;
        green  = !i_down;
    	break;

	case s_fault:			// Fatal errors
    case s_catch:
    case s_exit:
    	ctr += CFG_UPDATE_MS;
        button = toggle;
        if(ctr >= ((toggle) ? CFG_LED_FAULT_ON_MS : CFG_LED_FAULT_OFF_MS))
        {
	        ctr = 0;
	        toggle = !toggle;
        }
    	break;

    default:
    	break;
	}
    SETLEDBTN(button);		// Set the LEDs
	SETLEDRED(red);
    SETLEDGRN(green);
}

//--Monitors-------------------------------------------------------------------
//! On-demand monitoring actions
//-----------------------------------------------------------------------------
//! If the user presses a key, display special values
//! 
//-----------------------------------------------------------------------------
static void Monitors(void)
{
	int i;
    
	switch(SerialGetRx())
    {
    	case '\0':			// No input
        	break;
        default:			// Unknown input
        	outc('?');
            break;
        case 'A':			// Show ADC readings
        case 'a':
    		outc('\n');
        	for(i=0; i<NADCS; i++)
            {
            	outn(RawAdcs[i], 5, 10);
            }
        	break;
    }
}

//--main-----------------------------------------------------------------------
//! Mainline function
//-----------------------------------------------------------------------------
//! Initialise everything, then loop, displaying log data. When the ADC interrupt
//! sets its flag, run the main state machine.
//! 
//-----------------------------------------------------------------------------
int main(void)		// Elementary test of the LEARN action	
{
    StartHardware();

    ClearScreen();                  // Initial sign-on to debug screen
    outs_P(PSTR(CFG_SIGNON));       // Sign-on message, (in program flash)

    cli;                            // Arm the internal watchdog
    _WDR();
    WDTCR |= BV(WDCE) | BV(WDE);    // Watchdog safety
    WDTCR  = BV(WDE)  | 2;          // Armed, timeout = 65mS
    sei;

	for(;;)
    {
        if(timerflag)
        {                       // Time to run the main state machine?
            SETWATCHDOG(0);         // Reset the external watchdog line
            _WDR();                 // Reset the internal watchdog
            timerflag = FALSE;
            ExternalInputs();       // Filter the switch inputs
            DriveLeds();			// Set the LEDs
            StateMachine();         // Main state engine 
        }
        Monitors();
    }
    return 0;                   // Never reached
}
