//WARNING comment syntax is important,  if unsure, check out http://www.doxygen.org/manual.html

//---------------------------------------------------------------------------
//! \file serial.c
//!
//! \brief  Asynchronous serial comms. handler
//!
//! Interrupt driven serial TX
//!     Version:    1.01         
//!     Processor:  Atmel ATMega8
//!     Compiler:   GCC          
//!
//! \author David R Brooks
//!
//! \version 13/07/2007 DRB inital design and coding
//!
//! \todo                Receiver functions
//---------------------------------------------------------------------------
// 2007.09.29 11:05  [YYYY.MM.DD] Last edited (CodeWright)

//-Included files -----------------------------------------------------------
#include "serial.h"
#include "interrupt.h"

//-Local Symbols ------------------------------------------------------------
// [insert local #defines and variable types]

//-Globals ------------------------------------------------------------------
// [insert global variables]

//-Locals -------------------------------------------------------------------
static uint1 TXQ[32];
static uint1 TQpush;
static uint1 TQpull;
static bool1 TQactive;

//-Local prototypes ---------------------------------------------------------
//! \cond SEE_PROTOTYPE

static bool1 IsTQempty(uint1* ppt);
static void  ToTQ(uint1 c);
static uint1 FmTQ(void);
static bool1 TxChar(void);

//! \endcond

//-Local Functions ----------------------------------------------------------

//--TX interrupt-------------------------------------------------------------
//! Serial TX interrupt
//---------------------------------------------------------------------------
//! This interrupt occurs when the TX buffer becomes free, ie a char. has been
//! moved to the shift register.
//! 
//---------------------------------------------------------------------------
SIGNAL(SIG_UART_TRANS)
{
	(void)TxChar();
}

//--TX sender----------------------------------------------------------------
//! Pull a char. from the queue, and send.
//---------------------------------------------------------------------------
//! This is called both from the TX interrupt, and from the FIFO driver (to
//! start a new transmission).
//!
//! \return      TRUE if a char. was sent, FALSE if queue was empty
//!
//! \note        Run with ints. disabled
//! 
//---------------------------------------------------------------------------
static bool1 TxChar(void)
{
	uint1 c = FmTQ();			// Next char. from main FIFO
	
	if(c)
	{
		UDR = c;				// Send to TX
		return TRUE;
	}
	else
	{
		TQactive = FALSE;		// Nothing to send
		return FALSE;
	}
}

//---------------------------------------------------------------------------
//! Return TRUE if the FIFO is/will become empty
//---------------------------------------------------------------------------
//!
//! \param ppt             Addr. of value to use as push pointer
//! \return                TRUE if empty
//! 
//---------------------------------------------------------------------------
static bool1 IsTQempty(uint1* ppt)
{
	if(ppt)
	{
		return(*ppt == TQpull);
	}
	else
	{
		return(TQpush == TQpull);
	}
}

//--TQsucc-------------------------------------------------------------------
//! TX FIFO pointer's successor function
//---------------------------------------------------------------------------
//! Get next FIFO locn.
//! 
//! \param ptr          Current ptr.
//!
//! \note				sizeof(TXQ) must be a power of 2
//! 
//---------------------------------------------------------------------------
static inline uint1 TQsucc(uint1 ptr)
{
	ptr++;
	ptr &= (sizeof(TXQ)-1);
	return ptr;
}

//--ToTQ---------------------------------------------------------------------
//! Put a char. in the queue
//---------------------------------------------------------------------------
//! If the queue is full, skip this char.
//! 
//! \param c              Char. to load
//!
//---------------------------------------------------------------------------
static void ToTQ(uint1 c)
{
	uint1 pt = TQsucc(TQpush);
	
	while(IsTQempty(&pt))			// Wait, if no room in the queue?
		;

	TXQ[TQpush] = c;				// Load the char.
	TQpush = pt;
}

//--FmTQ---------------------------------------------------------------------
//! Get a char. from the FIFO, if any
//---------------------------------------------------------------------------
//! If the queue is empty, return NUL
//! 
//! \return               Returned char, or NUL
//!
//---------------------------------------------------------------------------
static uint1 FmTQ(void)
{
	if(IsTQempty(NULL))
	{
		return 0;			// FIFO is empty
	}
	else
	{
		uint1 c = TXQ[TQpull];
		TQpull  = TQsucc(TQpull);
		return c;
	}
}

//-Global Functions ---------------------------------------------------------

//--CommsInit----------------------------------------------------------------
//! Initialise UART (async. use)
//---------------------------------------------------------------------------
//! Start the comms. port, for TX interrupt, RX polled
//! 
//! \param rate_u4        Speed setting (bits/sec)
//!
//! \note                 Assumes 8MHz clock, and double-speed UART
//!
//---------------------------------------------------------------------------
bool1 CommsInit(uint4 rate_u4)
{
    uint2 baud_u2;                      // Baud-rate divisor
    int   i;

    const struct                        // Baud-rate table (8MHz oscillator, double speed)
    {
        uint4   speed_u4;
        uint2   factor_u2;
    } BaudTab[] = {
        {  2400, 416 },
        {  4800, 207 },
        {  9600, 103 },
        { 14400,  68 },
        { 19200,  51 },
        { 38400,  25 },
        { 57600,  16 },
        {115200,   8 }
    };

    for(i=0; i<DIMOF(BaudTab); i++)     // Table lookup
    {
        if(rate_u4 == BaudTab[i].speed_u4)
        {
            baud_u2 = BaudTab[i].factor_u2;
            UDR   = 0;                                  // Clear data register
            UCSRA = BV(U2X);                            // Double clock speed
            UCSRB = BV(RXEN)  | BV(TXEN)  | BV(TXCIE);  // RX & TX on (full duplex)
            UCSRC = BV(URSEL) | BV(UCSZ1) | BV(UCSZ0);  // Mode: 8N1
            UBRRH = baud_u2 >> 8;                       // Baud rate: 16 bits
            UBRRL = baud_u2 & 0x00ff;
            return TRUE;
        }
    }
    return FALSE;
}

//--outn-----------------------------------------------------------------------
//! Output n as an unsigned string, radix r
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param value      Value to output
//! \param width      Width to display
//! \param radix      Radix
//! 
//-----------------------------------------------------------------------------
void outn(int value, int width, int radix)
{
	PGM_P t = PSTR("0123456789ABCDEF");
	
    if(value<0)
    {
        outc('-');
        outn(-value, width-1, radix);
    }
    else
    {
        int q = value / radix;
        int r = value % radix;
        if(q > 0)
        {
            outn(q, width-1, radix);
			width = 0;
        }
		for(; width>0; width--)
		{
			outc(' ');
		}
        outc(pgm_read_byte(&t[r]));
    }
}

//--outc---------------------------------------------------------------------
//! Basic char. output on UART
//---------------------------------------------------------------------------
//! Interrupt driven
//! 
//! \param c          Char. to send
//! 
//! \pre              Assumes port is already set up (see CommsPortInit)
//! \note             LF is expanded to CR/LF
//! \sa               CommsPortInit()
//! 
//---------------------------------------------------------------------------
void outc(char c)
{
    if(c == '\n')                   // LF -> CR-LF
        outc('\r');
		
	ToTQ(c);						// Waits if queue is full

	cli;                            // Disable, in case the FIFO was emptied just here
	if(!TQactive)					// Need to start it?
	{
		TQactive = TxChar();		// Try to start the TX
	}
	sei;
}

//--outsw----------------------------------------------------------------------
//! Output string, pad to width
//-----------------------------------------------------------------------------
//! [insert detailed description]
//!
//! \param p          Ptr. to string
//! \param w          Minimum width
//!
//-----------------------------------------------------------------------------
void outsw(char* p, sint2 w)
{
    for(w -= outs(p); w > 0; w--)
    {
        outc(' ');
    }
}

//--outsw_P--------------------------------------------------------------------
//! Output string, pad to width (from flash)
//-----------------------------------------------------------------------------
//! [insert detailed description]
//!
//! \param p          Ptr. to string
//! \param w          Minimum width
//!
//-----------------------------------------------------------------------------
void  outsw_P(PGM_P p, sint2 w)
{
    for(w -= outs_P(p); w > 0; w--)
    {
        outc(' ');
    }
}

//--outs---------------------------------------------------------------------
//! Output a string, return its width on screen
//---------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \param p          Ptr. to string
//! \return           Displayed width of string
//! 
//---------------------------------------------------------------------------
uint1 outs(char* p)
{
    uint1 l = 0;

	while(*p)
	{
        if(*p >= ' ')
        {                   // Count of displayable chars.
            l++;
        }
		outc(*p++);
	}
    return l;
}

//--outs_P---------------------------------------------------------------------
//! Output a string from code flash
//-----------------------------------------------------------------------------
//! Similar to outs(), but accepts a pointer into code flash, not RAM.
//! Uses the std. library to access code flash as data.
//! 
//! \param p          String ptr. (in code flash)
//! \return           Displayed width of string
//! 
//! \note             Strings must be defined using prog_char
//-----------------------------------------------------------------------------
uint1 outs_P(PGM_P p)
{
    char c;
    uint1 l = 0;

	while((c = pgm_read_byte(p++)))
	{
        if(c >= ' ')
        {                   // Count of displayable chars.
            l++;
        }
		outc(c);
	}
    return l;
}

//--wait---------------------------------------------------------------------
//! Simple NuS delay
//---------------------------------------------------------------------------
//! A simple delay loop, about 1uS per count
//! 
//! \param delay      Delay in uS
//! 
//---------------------------------------------------------------------------
void wait(uint4 delay_us)
{
    while(delay_us--)
	{
        __asm__ __volatile__ ("nop" ::);
	}
}

//--WaitForTx----------------------------------------------------------------
//! Wait until TX FIFO is empty
//---------------------------------------------------------------------------
//! 
//---------------------------------------------------------------------------
void WaitForTx(void)
{
	while(!IsTQempty(NULL))
    {
	    ;
    }
}

//--IsTxqEmpty---------------------------------------------------------------
//! Check if the TX FIFO is empty
//---------------------------------------------------------------------------
//! 
//! \return        TRUE is it is empty
//!
//---------------------------------------------------------------------------
bool1 IsTxqEmpty(void)
{
	return IsTQempty(NULL);
}

//--SerialGetRx----------------------------------------------------------------
//! Return polled keyboard char. (or \0)
//-----------------------------------------------------------------------------
//! [insert detailed description]
//! 
//! \return           Returned char.
//! 
//-----------------------------------------------------------------------------
uint1 SerialGetRx(void)
{
    if(UCSRA & BV(RXC))
    {
        return(UDR);
    }
    else
    {
        return '\0';
    }
}
